/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

sunrise-set.c is for calculating the time of sunrise and sunset for
the current date and fixed latitude and longitude.

The program is based on rscalc.c by Jarmo Lammi, with suggestions by
David Smith.

$Revision: 0.0 $
$Date: 2007/01/18 19:36:00 $
$Id: sunrise-set.c,v 0.0 2007/01/18 19:36:00 john Exp $
$Log: sunrise-set.c,v $
Revision 0.0  2007/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

static char rcsid[] = "$Id: sunrise-set.c,v 0.0 2007/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Calculate sunrise and sunset\n",
    "Usage: sunrise-set [-d] [-L longitude] [-l latitude] [-r] [-s] [-v]\n",
    "    -d, print nothing, but return 0 if daytime, 2 if not\n",
    "    -L longitude, decimal local longitude, (-121.975916667)\n",
    "    -l latitude, decimal local latitude, (37.276833333)\n",
    "    -r, print actual sunrise\n"
    "    -s, print actual sunset\n"
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Not Daylight\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define NOTDAYLIGHT 2

#ifdef __STDC__

static double FNday (int y, int m, int d, float h); /* get the days to J2000 */
static double FNrange (double x); /* angle in the range 0 to 2 * pi */
static double f0 (double lat, double declin); /* calculate the hour angle */
static double f1 (double lat, double declin); /* calculate the hour angle for twilight times */
static double FNsun (double d); /* ecliptic longitude of the sun */
static char *showhrmn (double dhr); /* display decimal hours in hours and minutes */
static void print_message (int retval); /* print any error messages */

#else

static double FNday (); /* get the days to J2000 */
static double FNrange (); /* angle in the range 0 to 2 * pi */
static double f0 ();  /* calculate the hour angle */
static double f1 (); /* calculate the hour angle for twilight times */
static double FNsun (); /* ecliptic longitude of the sun */
static char *showhrmn (); /* display decimal hours in hours and minutes */
static void print_message (); /* print any error messages */

#endif

static double pi = (double) 3.14159265358979323846,
       degs,
       rads,
       L,
       g,
       daylen,
       SunDia = (double) 0.53, /* Sun radius degrees */
       AirRefr = (double) 34.0 / (double) 60.0; /* Atmospheric refraction degrees */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    int retval = NOERROR, /* return value, assume no error */
        c, /* command line switch */
        daytime = 0, /* print nothing, but return 0 if daytime, 2 if not, flag, 0 = false, 1 = true */
        sunrise = 0, /* print only actual sunrise, 0 = no, 1 = yes */
        sunset = 0; /* print only actual sunset, 0 = no, 1 = yes */

    double y,
        m,
        day,
        h,
        latit,
        longit,
        tzone,
        d,
        lambda,
        obliq,
        alpha,
        delta,
        LL,
        equation,
        ha,
        hb,
        twx,
        altmax,
        noont,
        settm,
        riset,
        hour,
        minutes,
        seconds,
        current;

    float inlat = (float) 37.276833333,
        inlon = (float) -121.975916667;

    time_t sekunnit;

    struct tm *p;

    while ((c = getopt (argc, argv, "dL:l:rst:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for print nothing, but return 0 if daytime, 2 if not, flag, 0 = false, 1 = true */

                daytime = 1; /* yes, set the print nothing, but return 0 if daytime, 2 if not, flag, 0 = false, 1 = true */
                break;

            case 'L': /* request for longitude? */

                inlon = atof (optarg); /* yes, get the longitude */
                break;

            case 'l': /* request for latitude? */

                inlat = atof (optarg); /* yes, get the latitude */
                break;

            case 'r': /* request for print only actual sunrise */

                sunrise = 1; /* set the print only actual sunrise, 0 = no, 1 = yes */
                break;

            case 's': /* request for print only actual sunset */

                sunset = 1; /* set the print only actual sunset, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {

        degs = (double) 180.0 / pi;
        rads = pi / (double) 180.0;

        /*

        get the date and time from the user
        read system date and extract the year

        */

        time (&sekunnit);

        p = localtime (&sekunnit);

        y = p->tm_year;
        y+= 1900;
        m = p->tm_mon + 1;
        day = p->tm_mday;
        hour = p->tm_hour;
        minutes = p->tm_min;
        seconds = p->tm_sec;
        tzone = (double) (p->tm_gmtoff / 3600);

        h = 12;

        latit = (double) inlat;
        longit = (double)inlon;

        d = FNday (y, m, day, h);

        /* use FNsun to find the ecliptic longitude of the sun */

        lambda = FNsun (d);

        /* obliquity of the ecliptic */

        obliq = (double) 23.439 * rads - (double) 0.0000004 * rads * d;

        /* find the RA and DEC of the sun */

        alpha = atan2 (cos (obliq) * sin (lambda), cos (lambda));
        delta = asin (sin (obliq) * sin (lambda));

        /* find the equation of time in minutes, correction suggested by David Smith */

        LL = L - alpha;

        if (L < pi)
        {
            LL += (double) 2.0 * pi;
        }

        equation = (double) 1440.0 * ((double) 1.0 - LL / pi / (double) 2.0);
        ha = f0 (latit, delta);
        hb = f1 (latit, delta);
        twx = hb - ha;  /* length of twilight in radians */
        twx = (double) 12.0 * twx / pi; /* length of twilight in hours */
        /* (void) printf("ha= %.2f   hb= %.2f \n", ha, hb); */

        /* conversion of angle to hours and minutes */

        daylen = degs *ha / (double) 7.5;

        if (daylen < (double) 0.0001)
        {
            daylen = (double) 0.0;
        }

        /* arctic winter */

        riset = (double) 12.0 - (double) 12.0 * ha / pi + tzone - longit / (double) 15.0 + equation / (double) 60.0;
        settm = (double) 12.0 + (double) 12.0 * ha / pi + tzone - longit / (double) 15.0 + equation / (double) 60.0;
        noont = riset + (double) 12.0 * ha / pi;
        altmax = (double) 90.0 + delta * degs - latit;

        /*

        correction for S HS suggested by David Smith to express altitude
        as degrees from the N horizon

        */

        if (latit < delta * degs)
        {
            altmax = (double) 180.0 - altmax;
        }

        if (riset > (double) 24.0)
        {
            riset-= (double) 24.0;
            noont = riset + (double) 12.0 * ha / pi; /* J. Conover, Sep. 22, 2013 */
        }

        if (settm > (double) 24.0)
        {
            settm-= (double) 24.0;
        }

        if (daytime == 0) /* print nothing, but return 0 if daytime, 2 if not, flag set, 0 = false, 1 = true? */
        {

            if ((sunrise == 0) && (sunset == 0)) /* either print only actual sunrise and print only actual sunset flags set? */
            {
                (void) printf ("Year           : %d \n", (int) y);
                (void) printf ("Month          : %d \n", (int) m);
                (void) printf ("Day            : %d \n", (int) day);
                (void) printf ("Days since Y2K : %d \n", (int) d);
                (void) printf ("Latitude       : %3.1f, longitude: %3.1f, timezone: %3.1f \n", (float) latit, (float) longit, (float) tzone);
                (void) printf ("Declination    : %.2f \n", delta * degs);
                (void) printf ("Daylength      : %s hours\n", showhrmn (daylen));
                (void) printf ("Sunrise        : %s\n", showhrmn (riset));

                /* amendment by David Smith */

                (void) printf ("Sun altitude   : %.2f degrees", altmax);
                (void) printf (latit >= (double) 0.0 ? " South" : " North");
                (void) printf (" at noontime = %s\n", showhrmn (noont));
                (void) printf ("Sunset         : %s\n", showhrmn (settm));
            }

            else
            {

                if (sunrise == 1) /* print only actual sunrise, 0 = no, 1 = yes? */
                {
                    (void) printf ("%s\n", showhrmn (riset)); /* print it */
                }

                if (sunset == 1) /* print only actual sunset, 0 = no, 1 = yes */
                {
                    (void) printf ("%s\n", showhrmn (settm)); /* print it */
                }

            }

        }

        else
        {
            current = hour + (minutes / (double) 60.0) + (seconds / (double) 3600.0); /* current time in hours, and fractions */

            if ((current < riset) || (current > settm))
            {
                retval = NOTDAYLIGHT; /* return value, assume not daylight */
            }

            /* (void) printf ("riset = %f, current = %f, settm = %f\n", riset, hour + (minutes / (double) 60.0) + (seconds / (double) 3600.0), settm); */
        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

static double FNday (int y, int m, int d, float h);

Get the days to J2000, h is UT in decimal hours.

FNday only works between 1901 to 2099 - see Meeus chapter 7

*/

#ifdef __STDC__

static double FNday (int y, int m, int d, float h)

#else

static double FNday (y, m, d, h)
int y;
int m;
int d;
float h;

#endif
{
    long int luku = - (long int) 7 * ((long int) y + ((long int) m + (long int) 9) / (long int) 12) / (long int) 4 + (long int) 275 * (long int) m / (long int) 9 + (long int) d;

    /*

    type casting necessary on PC DOS and TClite to avoid overflow

    */

    luku+= (long int) y * (long int) 367;
    return ((double) ((double) luku - (double) 730531.5 + (double) h / (double) 24.0));
}

/*

static double FNrange (double x);

returns an angle in the range 0 to 2 * pi

*/

#ifdef __STDC__

static double FNrange (double x)

#else

static double FNrange (x)
double x;

#endif

{
    double b = (double) 0.5 * x / pi,
           a = (double) 2.0 * pi * (b - (long) (b));

    if (a < 0)
    {
        a = (double) 2.0 * pi + a;
    }

    return (a);
}

/*

static double f0 (double lat, double declin);

calculate the hour angle

*/

#ifdef __STDC__

static double f0 (double lat, double declin)

#else

static double f0 (lat, declin)
double lat;
double declin;

#endif

{
    double fo,
           dfo;

    /* correction: different sign at S HS */

    dfo = rads * ((double) 0.5 * SunDia + AirRefr);

    if (lat < (double) 0.0)
    {
        dfo = -dfo;
    }

    fo = tan (declin + dfo) * tan (lat * rads);

    if (fo > (double) 0.99999)
    {
        fo = (double) 1.0; /* to avoid overflow */
    }

    fo = asin (fo) + pi / (double) 2.0;

    return (fo);
}

/*

static double f1 (double lat, double declin);

calculate the hour angle for twilight times

*/

#ifdef __STDC__

static double f1 (double lat, double declin)

#else

static double f1 (lat, declin)
double lat;
double declin;

#endif

{
    double fi,
           df1;

    /* correction: different sign at S HS */

    df1 = rads * (double) 6.0;

    if (lat < (double) 0.0)
    {
        df1 = -df1;
    }

    fi = tan (declin + df1) * tan (lat * rads);

    if (fi > (double) 0.99999)
    {
        fi = (double) 1.0; /* to avoid overflow */
    }

    fi = asin (fi) + pi / (double) 2.0;
    return (fi);
}

/*

static double FNsun (double d);

ecliptic longitude of the sun

*/

#ifdef __STDC__

static double FNsun (double d)

#else

static double FNsun (d)
double d;

#endif

{

    /* mean longitude of the sun */

    L = FNrange ((double) 280.461 * rads + (double) 0.9856474 * rads * d);

    /* mean anomaly of the sun */

    g = FNrange ((double) 357.528 * rads + (double) 0.9856003 * rads * d);

    /* ecliptic longitude of the sun */

    return (FNrange (L + (double) 1.915 * rads * sin (g) + (double) 0.02 * rads * sin ((double) 2.0 * g)));
}

/*

static char *showhrmn (double dhr);

decimal hours in hours and minutes conversion

returns a character reference to the conversion

*/

#ifdef __STDC__

static char *showhrmn (double dhr)

#else

static char *showhrmn (dhr)
double dhr;

#endif

{
    static char reg[BUFLEN + 1]; /* i/o buffer */

    int hr,
        mn;

    hr = (int) dhr;
    mn = (dhr - (double) hr) * (int) 60;

    (void) sprintf (reg, "%02d:%02d", hr, mn);

    return (reg); /* return a reference to the conversion */
}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
